/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: eval.c,v 1.1 2021-02-26 13:32:23-05 andrew Exp andrew $
 * 
 * Copyright (c) 2021 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "eval.h"
#include "lc.h"
#include "opcodes.h"
#include "symtab.h"
#include "util.h"
#include "psect.h"
#include "lexerstates.h"



extern int pass;


/*
 * FUNCTION	evalUnary
 * DESCRIPTION	Evaluate any unary operators (+,-,!,~) in token list 'tokv'
 *		and reduce token list accordingly.
 */

void evalUnary(Token **tokv)
{
  int i;
  int x = 0;
  int a;
  int tt;
  char buf[64];

  /* These operators evaluate right-to-left. */

  i = 0;
  /* Advance to end */
  while (tokv[i]) i++;
  do {
    i--;

    // Make sure we've evaluating an operator and not a number
    // if (tokIsNumber(tokv[i])) { i--; continue; }

    tt = tokv[i]->tok_type;
    // fprintf(stderr, "[%d]%d ", i, tt); fflush(stderr);
    if ((tt == PLUS) || (tt == MINUS) || (tt == LOGIC_NOT) || (tt == BIT_NOT)) {
      MESSAGE(MSG_DEBUG, "%d: evalUnary", i);

      // Skip if left token is a label or number
      if (i && tokIsLabel(tokv[i-1])) { i--; continue; }
      if (i && tokIsNumber(tokv[i-1])) { i--; continue; }

      MESSAGE(MSG_VERBOSE, "%d: Attempting evalUnary", i);

      // Must have argument on right
      if (tokv[i+1] == NULL) {
        fprintf(stderr, "Expression syntax error near '%s'\n", tokv[i]->tok_text);
        exit(1);
      }

      // Parse out the number
      // QUESTION: Is this still needed? Does evalString alreay
      //           parse numbers?
      a = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }
      if (tt == PLUS)  x = a;
      if (tt == MINUS) x = -a;
      if (tt == LOGIC_NOT) x = !a;
      if (tt == BIT_NOT)   x = ~a;
      FREE(tokv[i]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i]->tok_text = strdup(buf);

      tokShiftL(tokv, i+1, 1);
      // i--;
    }
  } while (i > 0);
  MESSAGE(MSG_DEBUG, "Leaving evalUnary()");
}


/*
 * FUNCTION	evalMulDiv
 * DESCRIPTION	Evaluate any multiply, divide, and modulus operators (*,/,%)
 *		in token list 'tokv' and reduce token list accordingly.
 */

void evalMulDiv(Token **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  i = 0;
  while (tokv[i]) {
    if ((tokv[i]->tok_type == TIMES) || (tokv[i]->tok_type == DIVIDE) || (tokv[i]->tok_type == MOD)) {
      // fprintf(stderr, "%d: Mult or div\n", i);
      // Evaluate multiplication or division
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }
      if (tokv[i]->tok_type == TIMES) x = a * b;
      if (tokv[i]->tok_type == DIVIDE) x = a / b;
      if (tokv[i]->tok_type == MOD) x = a % b;

      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


/*
 * FUNCTION	evalAddSub
 * DESCRIPTION	Evaluate any add and subtract operators (+,-)
 *		in token list 'tokv' and reduce token list accordingly.
 */

void evalAddSub(Token **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalAddSub: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    // Make sure we've evaluating an operator and not a number
    if (tokIsNumber(tokv[i])) { i++; continue; }

    // fprintf(stderr, "*** evalAddSub: evaluating token %d: '%s'\n", i, tokv[i]);
    if ((tokv[i]->tok_type == PLUS) || (tokv[i]->tok_type == MINUS)) {
      // fprintf(stderr, "%d: Add or sub\n", i);
      // Evaluate addition or subtraction
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }
      if (tokv[i]->tok_type == PLUS) x = a + b;
      if (tokv[i]->tok_type == MINUS) x = a - b;
      // fprintf(stderr, "Freeing left token (%s)\n", tokv[i-1]);
      assert(tokv[i-1]->tok_text);
      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      // fprintf(stderr, "Setting left token to result\n");
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      // fprintf(stderr, "Before tokShiftL:\n"); tokDump(tokv);
      tokShiftL(tokv, i, 2);
      // fprintf(stderr, "After tokShiftL:\n"); tokDump(tokv);
      i--;
    }
    i++;
  }
}


/*
 * FUNCTION	evalBitShift
 * DESCRIPTION	Evaluate any bit shift operators (<<,>>)
 *		in token list 'tokv' and reduce token list accordingly.
 */

void evalBitShift(Token **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalBitShift: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    if ((tokv[i]->tok_type == SHIFTLEFT) || (tokv[i]->tok_type == SHIFTRIGHT)) {
      // Evaluate addition or subtraction
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }
      if (tokv[i]->tok_type == SHIFTLEFT)  x = a << b;
      if (tokv[i]->tok_type == SHIFTRIGHT) x = a >> b;
      // fprintf(stderr, "Freeing left token (%s)\n", tokv[i-1]);
      assert(tokv[i-1]->tok_text);
      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


/*
 * FUNCTION	evalInequality
 * DESCRIPTION	Evaluate any inequality comparison operators (<,<=,>,>=)
 *		in token list 'tokv' and reduce token list accordingly.
 */

void evalInequality(Token **tokv)
{
  int i;
  int x = 0;
  int a,b;
  int tt;
  char buf[64];

  i = 0;
  while (tokv[i]) {
    tt = tokv[i]->tok_type;
    if ((tt == LESSTHAN) || (tt == LESSEQUAL)
        || (tt == GREATERTHAN) || (tt == GREATEREQUAL)) {
      // Evaluate inequalities
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }

      if (tokv[i]->tok_type == LESSTHAN)     x = (a < b);
      if (tokv[i]->tok_type == LESSEQUAL)    x = (a <= b);
      if (tokv[i]->tok_type == GREATERTHAN)  x = (a > b);
      if (tokv[i]->tok_type == GREATEREQUAL) x = (a >= b);
      assert(tokv[i-1]->tok_text);
      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }  /* while */
}


/*
 * FUNCTION	evalEquality
 * DESCRIPTION	Evaluate any equal and not-equal operators (==,!=)
 *		in token list 'tokv' and reduce token list accordingly.
 */

void evalEquality(Token **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  i = 0;
  while (tokv[i]) {
    if ((tokv[i]->tok_type == EQUAL) || (tokv[i]->tok_type == NOTEQUAL)) {
      // Evaluate Equal (==) and Not equal (!=)
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }
      if (tokv[i]->tok_type == EQUAL)    x = (a == b);
      if (tokv[i]->tok_type == NOTEQUAL) x = (a != b);
      assert(tokv[i-1]->tok_text);
      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


void evalBitAnd(Token **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalBitAnd: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    if (tokv[i]->tok_type == BIT_AND) {
      // Evaluate bitwise AND
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }
      if (tokv[i]->tok_type == BIT_AND) x = a & b;
      assert(tokv[i-1]->tok_text);
      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


void evalBitXor(Token **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalBitXor: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    if (tokv[i]->tok_type == BIT_XOR) {
      // Evaluate bitwise XOR
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }
      if (tokv[i]->tok_type == BIT_XOR) x = a ^ b;
      assert(tokv[i-1]->tok_text);
      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


void evalBitOr(Token **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalBitOr: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    if (tokv[i]->tok_type == BIT_OR) {
      // Evaluate bitwise OR
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }
      if (tokv[i]->tok_type == BIT_OR) x = a | b;
      assert(tokv[i-1]->tok_text);
      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


void evalByteSelect(Token **tokv)
{
  int i;
  int x = 0;
  int a;
  char buf[64];

  message(MSG_DEBUG, "Entering evalByteSelect()");
  i = 0;
  while (tokv[i]) {
    // fprintf(stderr, " [%d]%d", i, tokv[i]->tok_type); fflush(stderr);
    if ((tokv[i]->tok_type == LOBYTE) || (tokv[i]->tok_type == HIBYTE)) {
      // Must have numeric argument on left
      if ((i==0) || (!tokIsLabel(tokv[i-1]) && !tokIsNumber(tokv[i-1]))) {
        fprintf(stderr, "Expression syntax error near '%s'\n", tokv[i]->tok_text);
        exit(1);
      }

      // Parse out the number
      // QUESTION: Is this still needed? Does evalString alreay
      //           parse numbers?
      a = parseNumber(tokv[i-1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }

      if (tokv[i]->tok_type == HIBYTE) x = (a >> 8) & 0xff;
      if (tokv[i]->tok_type == LOBYTE) x = a & 0xff;
      FREE(tokv[i]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 1);
      i--;
    }  /* if LOBYTE|HIBYTE */
    i++;
  }  /* while */
}


void evalLogic(Token **tokv, const int token)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  i = 0;
  while (tokv[i]) {
    if (tokv[i]->tok_type == token) {
      /* Evaluate logic AND/XOR/OR */
      /* Need larg and rarg */
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }

      a = parseNumber(tokv[i-1]->tok_text);
      b = parseNumber(tokv[i+1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]->tok_text);
        exit(1);
      }

      if (tokv[i]->tok_type == LOGIC_AND) x = a && b;
      if (tokv[i]->tok_type == LOGIC_XOR) x = (a || b) && !(a && b);
      if (tokv[i]->tok_type == LOGIC_OR)  x = a || b;
      assert(tokv[i-1]->tok_text);
      FREE(tokv[i-1]->tok_text);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1]->tok_type = INTEGER;
      tokv[i-1]->tok_text = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }  /* if token match */
    i++;
  }
}


void evalTernary(Token **tokv)
{
  int i;
  int a;

  message(MSG_DEBUG, "Entering evalTernary()");
  i = 0;
  while (tokv[i]) {
    // fprintf(stderr, " [%d]%d", i, tokv[i]->tok_type); fflush(stderr);
    if (tokv[i]->tok_type == TERNARY) {
      message(MSG_VERBOSE, "Evaluating ternary expression at %d", i);
      // Evaluate ternary operator (?:)
      if ((i == 0) || (tokv[i+1] == NULL) || (tokv[i+2] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]->tok_text);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]->tok_text);
        exit(1);
      }

      /* Since this is the lowest precedence (apart from possible , operator)
       * we should be able to assume the form "a?b:c" with a,b,c being
       * single tokens.
       */
      if (a) {
        /* Use first value (between ? and :) */
        tokShiftL(tokv, i-1, 2);	/* Shift out "a?" */
        tokShiftL(tokv, i, 2);		/* Shift out ":c" */
      } else {
        /* Use second value (after :) */
        tokShiftL(tokv, i-1, 4);	/* Shift out "a?b:" */
      }
      i--;
    }  /* if TERNARY */
    i++;
  }  /* while */
}


extern void evalParens(Token **tokv);

/*
 * FUNCTION	evalOperators
 * DESCRIPTION	Evaluate all operators in token list 'tokv' and (hopefully)
 *		reduce list to a single numeric value.
 */

void evalOperators(Token **tokv)
{
  evalParens(tokv);
  evalUnary(tokv);
  evalMulDiv(tokv);
  evalAddSub(tokv);
  evalBitShift(tokv);
  evalInequality(tokv);
  evalEquality(tokv);
  evalBitAnd(tokv);
  evalBitXor(tokv);
  evalBitOr(tokv);
  evalByteSelect(tokv);
  evalLogic(tokv, LOGIC_AND);
  evalLogic(tokv, LOGIC_XOR);
  evalLogic(tokv, LOGIC_OR);
  evalTernary(tokv);
}


/*
 * FUNCTION	evalParens
 * DESCRIPTION	Evaluate any subexpressions in parentheses, calling
 *		evalOperators on innermost expressions first, until
 *		all parentheses have been reduced.
 */

void evalParens(Token **tokv)
{
  int i;
  int pstart = -1;
  int pdepth = 0;
  Token **subv;

  i=0;
  while (tokv[i]) {
    if (tokv[i]->tok_type == OPAREN) {
      // Begin parenthesis evaluation
      // fprintf(stderr, "----> Open Paren\n");
      if (pstart < 0) pstart = i;
      pdepth++;
    }
    if (tokv[i]->tok_type == CPAREN) {
      // End parenthesis evaluation
      // fprintf(stderr, "----> End Paren\n");
      if (pstart < 0) {
        fprintf(stderr, "ERROR: Syntax error: close paren without matching open paren.\n");
        exit(1);
      }
      pdepth--;
      if (pdepth == 0) {
        /* Evaluate from pstart to here */
        subv = tokDup(tokv, pstart+1, i-1);
        // fprintf(stderr, "(2) Sub-clause to evaluate:\n");
        // dumptokens(subv);
        evalOperators(subv);
        /* Expression should have evaluated down to one value, we hope. */
        tokv[pstart]->tok_type = subv[0]->tok_type;
        tokv[pstart]->tok_text = strdup(subv[0]->tok_text);
        tokShiftL(tokv, pstart+1, (i-(pstart)));
        FREE(subv[0]->tok_text);
        FREE(subv[0]);
        pstart = -1;    // Reset to no open paren
      }  // pdepth == 0
    }  // end paren
    i++;
  }
}


/*
 * FUNCTION	evalLabels
 * DESCRIPTION	Evaluate any labels in token list 'tokv' and substitute
 *		their value if known.
 * RETURNS	1 for success (for 1 token remaining), or EVAL_FAIL for
 *		failure.
 */

int evalLabels(Token **tokv)
{
  int i;
  int x;
  char xstr[32];

  /* Resolve labels */
  i=0;
  while (tokv[i]) {
    if (tokIsLabel(tokv[i])) {
      /* Is this line redundant?  Labels can't begin wtih digit. */
      if (!isdigit(tokv[i]->tok_text[0])) {
        // fprintf(stderr, "Resolving label or literal '%s'\n", tokv[i]->tok_text);
        if (!strcmp(tokv[i]->tok_text, "$")) {
          // special 'here' label
          MESSAGE(MSG_INFO, "evalString(): Evaluating $ to 0x%04x",
              (pass ? psAddrBase(curpsect->pnum) : 0) + lcGet());
          snprintf(xstr, sizeof(xstr), "%d",
              (pass ? psAddrBase(curpsect->pnum) : 0) + lcGet());
          FREE(tokv[i]->tok_text);
          tokv[i]->tok_text = strdup(xstr);
        } else {
          // Symbol to lookup
          x = symLookup(tokv[i]->tok_text);
          MESSAGE(MSG_INFO, "evalString(): Evaluating %s to 0x%04x",
              tokv[i]->tok_text, x);
          if (x != EVAL_FAIL) {
            snprintf(xstr, sizeof(xstr), "%d", x);
            FREE(tokv[i]->tok_text);
            tokv[i]->tok_text = strdup(xstr);
          } else {
            // If a symbol couldn't resolve, return error so it gets
            // marked as a forward reference to (hopefully) be resolved
            // during pass 2.
            //
            tokFree(tokv);
            return EVAL_FAIL;
          }  // if/else symLookup resolved symbol
        }  // if/else 'here' label
      }  // not number
    }
    i++;
  }  // while (label resolving)
  return 1;	/* Return 1 for success */
}


/*
 * FUNCTION	evalBraces
 * DESCRIPTION	Evaluate expressions inside braces ({,}) and concatenate
 *		result with left and right tokens.
 *
 *  Need to resolve any labels at some point: first? last? both?
 */

void evalBraces(Token **tokv)
{
  int i;
  int pstart = -1;
  int pdepth = 0;
  Token **subv;
  char rstr[256];	/* Result of concatenation */

  MESSAGE(MSG_DEBUG, "  Entering evalBraces()");
  i=0;
  while (tokv[i]) {
    if (tokv[i]->tok_type == OBRACE) {
      // Begin parenthesis evaluation
      // fprintf(stderr, "----> Open Brace\n");
      if (pstart < 0) pstart = i;
      pdepth++;
      MESSAGE(MSG_DEBUG, "  Found opening brace at %d, depth=%d, '%s'", pstart,pdepth, tokv[i]->tok_text);
    }
    if (tokv[i]->tok_type == CBRACE) {
      // End parenthesis evaluation
      // fprintf(stderr, "----> End Brace\n");
      if (pstart < 0) {
        fprintf(stderr, "ERROR: Syntax error: close brace without matching open brace.\n");
        exit(1);
      }
      pdepth--;
      MESSAGE(MSG_DEBUG, "  Found closing brace at %d, depth=%d, '%s'", i,pdepth, tokv[i]->tok_text);
      if (pdepth == 0) {
        /* Evaluate from pstart to here */
        message(MSG_INFO, "  Evaluating brace subexpression [%d..%d]", pstart+1, i-1);
        subv = tokDup(tokv, pstart+1, i-1);
        // fprintf(stderr, "(2) Sub-clause to evaluate:\n");
        // dumptokens(subv);

        message(MSG_DEBUG, "    First evalLabels()");
	evalLabels(subv);
        message(MSG_DEBUG, "    evalOperators()");
        evalOperators(subv);
        message(MSG_DEBUG, "    Second evalLabels()");
        evalLabels(subv);	/* Any possibility of operators forming */
				/* a new label?  (embedded {sym}?)      */

        /* Expression should have evaluated down to one value, we hope. */
        /* First, reduce subexpression to single value */
        tokv[pstart]->tok_type = subv[0]->tok_type;
        tokv[pstart]->tok_text = strdup(subv[0]->tok_text);
        tokShiftL(tokv, pstart+1, (i-(pstart)));
        FREE(subv[0]->tok_text);
        FREE(subv[0]);
        /* Now concatenate with left and right. */
        /* Right */
        if (tokv[pstart+1]) {
          snprintf(rstr, sizeof(rstr), "%s%s",
              tokv[pstart]->tok_text, tokv[pstart+1]->tok_text);
          tokv[pstart]->tok_type = STRING;
          FREE(tokv[pstart]->tok_text);
          tokv[pstart]->tok_text = strdup(rstr);
          tokShiftL(tokv, pstart+1, 1);
        }
        /* Left */
        if (pstart > 0) {
          snprintf(rstr, sizeof(rstr), "%s%s",
              tokv[pstart-1]->tok_text, tokv[pstart]->tok_text);
          tokv[pstart-1]->tok_type = STRING;
          FREE(tokv[pstart-1]->tok_text);
          tokv[pstart-1]->tok_text = strdup(rstr);
          tokShiftL(tokv, pstart, 1);
        }
        pstart = -1;    // Reset to no open paren
      }  // pdepth == 0
    }  // end paren
    i++;
  }
}


/*
 * FUNCTION	evalString
 * DESCRIPTION	Entry point for operand string evaluation.  Splits
 *		'str' into tokens in list 'tokv', then evalutate
 *		and reduces list to (hopefully) a single numeric value.
 * RETURNS	The final evaluated value of 'str' or EVAL_FAIL on failure.
 */

int evalString(const char *str)
{
  Token **tokv;

  MESSAGE(MSG_DEBUG, "Evaluating string '%s'", str);
  tokv = tokSplit(str);

  evalBraces(tokv);

  if (evalLabels(tokv) == EVAL_FAIL) return EVAL_FAIL;

  // fprintf(stderr, "Initial tokv[] array:\n");
  // tokDump(tokv);
  // printf("    Total tokens: %d\n", tokCount(tokv));
  evalOperators(tokv);
  // printf("After evals: %d\n", tokCount(tokv));
  // sscanf(tokv[0], "%i", &x);
  return parseNumber(tokv[0]->tok_text);
}


