/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: macro.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $
 * 
 * Copyright (c) 2017 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "macro.h"
#include "opcodes.h"
#include "util.h"



/* Hash of MACRO_SIZE macrolists */
struct macrolist *mactabs[MACRO_SIZE];


void macInit(void)
{
  int i;

  for (i=0; i<MACRO_SIZE; i++) {
    mactabs[i] = NULL;
  }
}


void macDestroy(void)
{
  int i;
  struct macrolist *mp;

  for (i=0; i<MACRO_SIZE; i++) {
    while (mactabs[i] != NULL) {
      mp = (mactabs[i])->next;
      free((mactabs[i])->label);
      stDestroy(mactabs[i]->s);
      free(mactabs[i]);
      mactabs[i] = mp;
    }
  }
}


StateList ** macAdd(const char *buf, unsigned int f)
{
  int i;
  struct macrolist *ment;

  if (!buf[0]) {
    fprintf(stderr, "WARNING: Attempt to add null macro name. Ignoring...\n");
    return NULL;
  }

  // fprintf(stderr, "     Pearson hash = %02x\n", Pearson_hash((void *)buf, strlen(buf)));
  i = (buf[0] % MACRO_SIZE);
  ment = (struct macrolist *)malloc(sizeof(struct macrolist));
  ment->label = strdup(buf);
  ment->s = stInit();
  ment->invocations = 0;
  ment->flags = f;
  ment->next = mactabs[i];
  mactabs[i] = ment;
  return ment->s;
}


StateList * macLookup(const char *lbl)
{
  int i;
  struct macrolist *sp;

  i = (lbl[0] % MACRO_SIZE);
  sp = mactabs[i];
  while (sp) {
    if (!strcmp(sp->label, lbl)) return *(sp->s);
    sp = sp->next;
  }
  return NULL;
}


int macInvocations(const char *lbl, const int increment)
{
  int i;
  struct macrolist *sp;

  i = (lbl[0] % MACRO_SIZE);
  sp = mactabs[i];
  while (sp) {
    if (!strcmp(sp->label, lbl)) {
      (sp->invocations) += increment;
      return sp->invocations;
    }
    sp = sp->next;
  }
  return -1;
}


// Reset macro state (i.e. invocation counts) for next pass
//
void macReset(void)
{
  int i;
  struct macrolist *mp;

  for (i=0; i<MACRO_SIZE; i++) {
    mp = mactabs[i];
    while (mp) {
      mp->invocations = 0;
      mp = mp->next;
    }  // while
  }  // for
}


void macDump(FILE *lfp)
{
  int i;
  struct macrolist *sp;

  for (i=0; i<MACRO_SIZE; i++) {
    sp = mactabs[i];
    while (sp != NULL) {
      fprintf(lfp, "\tmacro %s (%d %04x)\n", sp->label, sp->invocations, sp->flags);
      stDump(sp->s, lfp);
      fprintf(lfp, "\tendm %s\n", sp->label);
      sp = sp->next;
    }
  }
}


