/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: symtab.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $
 * 
 * Copyright (c) 2017-2021 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "eval.h"
#include "lc.h"
#include "symtab.h"
#include "opcodes.h"
#include "psect.h"
#include "source.h"
#include "util.h"



/* Hash of SYMTAB_SIZE symlists */
struct symlist *symtabs[SYMTAB_SIZE];

int allglobal = 1; /* All symbols (going forward) will be declared global */

struct symlist *lastsym = NULL;


void symInit(void)
{
  int i;

  for (i=0; i<SYMTAB_SIZE; i++) symtabs[i] = NULL;
}


void symDestroy(void)
{
  int i;
  struct symlist *sp;

  for (i=0; i<SYMTAB_SIZE; i++) {
    while (symtabs[i] != NULL) {
      sp = (symtabs[i])->next;
      FREE((symtabs[i])->label);
      FREE((symtabs[i])->file);
      FREE((symtabs[i])->pgroup);
      FREE(symtabs[i]);
      symtabs[i] = sp;
    }
  }
}


int symInScope(const struct symlist *sp, const char *lbl)
{
  /* Check that symbol is in scope */
  if (strcmp(sp->label, lbl)) return 0;

  if (!(sp->flags & SF_SCOPE_GLOBAL)) {
    if (strcmp(sp->file, srcname)) return 0;

    if (!(sp->flags & SF_SCOPE_FILE)) {
      if ((sp->pgroup) && (!strcmp(sp->pgroup, curpsect->pgroup))) return 1;
      if (sp->psect != curpsect->pnum) return 0;
    }  /* if !file */
  }  /* if !global */
  return 1;
}


void symAdd(const char *buf, const int loc, unsigned int f)
{
  int i;
  struct symlist *sp;
  struct symlist *sent;

  if (!buf[0]) {
    fprintf(stderr, "WARNING: Attempt to add null symbol. Ignoring...\n");
    return;
  }

  // fprintf(stderr, "%04x symAdd: Adding symbol '%s' = %04x, flags = %04x\n",
  //     lc, buf, loc, f);
  // fprintf(stderr, "     Pearson hash = %02x\n", Pearson_hash((void *)buf, strlen(buf)));
  i = Pearson_hash((void *)buf, strlen(buf)) % SYMTAB_SIZE;
  /* First, see if there's an existing (possibly undeclared) entry */
  sp = symtabs[i];
  while (sp) {
    if (symInScope(sp, buf)) {
      /* Here, the label, psect, and file are all in scope */
      /* If not declared, modify this entry; else throw warning/error */
      if (sp->flags & SF_DECLARED) {
        message(MSG_WARNING, "Attempt to redeclare symbol '%s'", buf);
        return;
      }
      message(MSG_INFO, "Declaring symbol '%s' with existing scope", buf);
      message(MSG_INFO, "  Initial flags: %04x", sp->flags);
      /* 'label', 'file', and 'psect' elements already set */
      sp->val = loc;
      sp->size = 0;
      sp->flags |= (f | SF_DECLARED);
      /* Global scope implies file scope */
      if (sp->flags & SF_SCOPE_GLOBAL) sp->flags |= SF_SCOPE_FILE;
      message(MSG_INFO, "  Final flags: %04x", sp->flags);
      if (!(sp->flags & SF_ABS)) { symSetSize(loc); lastsym = sp; }
      return;
    }
    sp = sp->next;
  }

  // sent = (struct symlist *)malloc(sizeof(struct symlist));
  sent = NEW(struct symlist);
  sent->label = strdup(buf);
  sent->file = strdup(srcname);
  sent->pgroup = curpsect->pgroup ? strdup(curpsect->pgroup) : NULL;
  sent->val = loc;
  sent->size = 0;
  sent->psect = curpsect->pnum;
  sent->flags = f | SF_DECLARED;
  if (allglobal) sent->flags |= SF_SCOPE_GLOBAL;
  /* Global scope implies file scope */
  if (sent->flags & SF_SCOPE_GLOBAL) sent->flags |= SF_SCOPE_FILE;
  sent->next = symtabs[i];
  symtabs[i] = sent;
  if (!(sent->flags & SF_ABS)) { symSetSize(loc); lastsym = sent; }
}


int symModify(const char *lbl, const int loc, unsigned int f)
{
  int i;
  struct symlist *sp;

  i = Pearson_hash((void *)lbl, strlen(lbl)) % SYMTAB_SIZE;
  // i = (lbl[0] % SYMTAB_SIZE);
  sp = symtabs[i];
  while (sp) {
    /* Check that symbol is in scope */
    if (symInScope(sp, lbl)) {

      /* Check for read-only status */
      if ((sp->flags & SF_DECLARED) && (!(sp->flags & SF_MOD)))
          return EVAL_FAIL;

      /* Either not declared or modifiable */
      /* Okay to modify */
      sp->val = loc;
      sp->size = 0;
      sp->psect = curpsect->pnum;
      sp->flags = f | SF_DECLARED;
      if (allglobal) sp->flags |= SF_SCOPE_GLOBAL;
      /* Global scope implies file scope */
      if (sp->flags & SF_SCOPE_GLOBAL) sp->flags |= SF_SCOPE_FILE;
      if (!(sp->flags & SF_ABS)) { symSetSize(loc); lastsym = sp; }
      return loc;
    }
    sp = sp->next;
  }
  /* Modifying undefined symbol.  Define it. */
  symAdd(lbl, loc, f);
  return loc;
}


inline int symValue(const struct symlist * sp)
{
  return ((sp->flags & SF_ABS) ? 0 : psAddrBase(sp->psect)) + sp->val;
}


/* Look up a symbol with a specified scope */
int symLookupScope(const char *lbl, int scope)
{
  int i;
  struct symlist *sp;

  i = Pearson_hash((void *)lbl, strlen(lbl)) % SYMTAB_SIZE;
  sp = symtabs[i];
  while (sp) {
    if ((!strcmp(sp->label, lbl)) && (sp->flags & SF_DECLARED)) {
      message(MSG_INFO, "Found label match for '%s'", lbl);
      switch (scope) {
        case SF_SCOPE_LOCAL:
          if (!(sp->flags & (SF_SCOPE_FILE | SF_SCOPE_GLOBAL))) {
              if ((sp->pgroup && curpsect->pgroup) && (!strcmp(sp->pgroup, curpsect->pgroup)))
                  return symValue(sp);
              if (sp->psect == curpsect->pnum) return symValue(sp);
          }
          break;
        case SF_SCOPE_FILE:
          message(MSG_INFO, "  File Scope, flags=%04x, file='%s'", sp->flags, sp->file);
          if (((sp->flags & SF_SCOPE_FILE) && !(sp->flags & SF_SCOPE_GLOBAL)) &&
              (!strcmp(sp->file, srcname)))
              return symValue(sp);
          break;
        case SF_SCOPE_GLOBAL:
          if (sp->flags & scope)
              return symValue(sp);
          break;
        default:	/* Illegal scope value */
          message(MSG_CRITICAL, "symLookupScope called with illegal scope %d", scope);
          return EVAL_FAIL;
      }  /* switch */
    }  /* if label match and declared */
    sp = sp->next;
  }
  return EVAL_FAIL;
}


int symLookup(const char *lbl)
{
  int val;

  /* Look for a local symbol first, then file scope, and finally global */
  if ((val = symLookupScope(lbl, SF_SCOPE_LOCAL )) != EVAL_FAIL) return val;
  if ((val = symLookupScope(lbl, SF_SCOPE_FILE  )) != EVAL_FAIL) return val;
  if ((val = symLookupScope(lbl, SF_SCOPE_GLOBAL)) != EVAL_FAIL) return val;
  return EVAL_FAIL;
}


void symSetScope(const char *lbl, const int scope)
{
  int i;
  struct symlist *sp;
  struct symlist *sent;

  i = Pearson_hash((void *)lbl, strlen(lbl)) % SYMTAB_SIZE;
  sp = symtabs[i];
  while (sp) {
    // Should be check for same file as well, or does same psect imply this?
    if ((!strcmp(sp->label, lbl)) && (sp->psect == curpsect->pnum)) {
      sp->flags |= (scope & (SF_SCOPE_FILE | SF_SCOPE_GLOBAL));
      /* Global scope implies file scope */
      if (sp->flags & SF_SCOPE_GLOBAL) sp->flags |= SF_SCOPE_FILE;
      return;
    }
    sp = sp->next;
  }

  /* Add entry for not-yet-declared symbol */
  // sent = (struct symlist *)malloc(sizeof(struct symlist));
  sent = NEW(struct symlist);
  sent->label = strdup(lbl);
  sent->file = strdup(srcname);
  sent->pgroup = curpsect->pgroup ? strdup(curpsect->pgroup) : NULL;
  sent->val = -1;	/* Symbol value not yet defined */
  sent->size = 0;
  sent->psect = curpsect->pnum;
  sent->flags = scope;	/* No flags known except scope */
  if (allglobal) sent->flags |= SF_SCOPE_GLOBAL;
  /* Global scope implies file scope */
  if (sent->flags & SF_SCOPE_GLOBAL) sent->flags |= SF_SCOPE_FILE;
  sent->next = symtabs[i];
  symtabs[i] = sent;
}


void symSetSize(const int loc)
{
  if (lastsym == NULL) return;

  lastsym->size = loc - lastsym->val;
  lastsym = NULL;
}


int symFlags(const char *lbl)
{
  int i;
  struct symlist *sp;

  i = Pearson_hash((void *)lbl, strlen(lbl)) % SYMTAB_SIZE;
  // i = (lbl[0] % SYMTAB_SIZE);
  sp = symtabs[i];
  while (sp) {
    if (!strcmp(sp->label, lbl)) {
      MESSAGE(MSG_DEBUG, "symFlags(): Found symbol '%s'", sp->label);
      return sp->flags;
    }
    sp = sp->next;
  }
  return EVAL_FAIL;
}


void symDumpFlags(FILE *lfp, const int f)
{
  fprintf(lfp, "< ");
  if (f & SF_FORWARD) fprintf(lfp, "FORWARD ");
  if (f & SF_ABS)     fprintf(lfp, "ABS ");
  if (f & SF_MOD)     fprintf(lfp, "MODIFIABLE ");
  if (f & SF_SCOPE_FILE)    fprintf(lfp, "SCOPE_FILE ");
  if (f & SF_SCOPE_GLOBAL)  fprintf(lfp, "SCOPE_GLOBAL ");
  if (f & SF_DECLARED)      fprintf(lfp, "DECLARED ");
  fprintf(lfp, ">");
}


void symDump(FILE *lfp)
{
  int i;
  int n;
  int nmax = 0;
  struct symlist *sp;

  fprintf(lfp, "; label@psect (file:pgroup) value size flags\n");
  for (i=0; i<SYMTAB_SIZE; i++) {
    sp = symtabs[i];
    n = 0;
    while (sp != NULL) {
      fprintf(lfp, "%s@%d (%s:%s) %04x %04x %04x", sp->label, sp->psect, sp->file, sp->pgroup ? sp->pgroup : "", sp->val, sp->size, sp->flags);
      symDumpFlags(lfp, sp->flags);
      fprintf(lfp, "\n");
      sp = sp->next;
      n++;
    }
    fprintf(lfp, "; Bucket %d list depth: %d\n", i, n);
    if (n > nmax) nmax = n;
  }
  fprintf(lfp, "; Max list depth: %d\n", nmax);
}


int symResolveAddr(const char *op)
{
  int opa;
  char alabel[64];

  // Evaluate the string. If not error condition, return value.
  opa = evalString(op);
  if (opa != EVAL_FAIL) return opa;

  // Really, the rest of this should now be redundant.
  MESSAGE(MSG_VERBOSE, "symResolveAddr: evalString(\"%s\") failed. Why?", op);

  // Numeric operand
  //
  if ( isdigit(op[0]) || ( (tolower(op[0]) == 'b') && (op[1] == '\''))) {
    // sscanf(op, "%i", &opa);
    opa = parseNumber(op);
    if (opa != EVAL_FAIL) return opa;
  }

  // Label operand
  //
  int i=0;
  while (op[i] && (op[i]!=' ') && (op[i]!='\t')) {
    alabel[i] = op[i];
    i++;
  }
  alabel[i] = '\0';

  // Special "current location counter" symbol '$'
  //
  if (!strcmp(alabel, "$")) {
    message(MSG_NOTICE, "symResolveAddr(): $ = 0x%04x", lcGet());
    return lcGet();
  }

  opa = symLookup(alabel);
  if (opa == EVAL_FAIL) {
    return EVAL_FAIL;
  }
  return opa;
}


int symResolveLiteral(const char *op)
{
  int opa;
  char alabel[64];

  MESSAGE(MSG_DEBUG, "Entering symResolveLiteral; operand = '%s'", op);

  // Evaluate the string. If not error condition, return value.
  opa = evalString(op);
  if (opa != EVAL_FAIL) return opa;

  message(MSG_NOTICE, "evalString() failed resolving");

  // Numeric operand
  if ( isdigit(op[0]) || ( (tolower(op[0]) == 'b') && (op[1] == '\''))) {
    // sscanf(op, "%i", &opa);
    opa = parseNumber(op);
    if (opa != EVAL_FAIL) return opa;
  }

  // Constant operand
  int i=0;
  while (op[i] && (op[i]!=' ') && (op[i]!='\t')) {
    alabel[i] = op[i];
    i++;
  }
  alabel[i] = '\0';

  if (!strcmp(alabel, "$")) {
    message(MSG_NOTICE, "symResolveLiteral(): $ = 0x%04x", lcGet());
    return lcGet();
  }
  MESSAGE(MSG_DEBUG, "symResolveLiteral: Looking up symbol %s\n", alabel);
  opa = symLookup(alabel);
  // opa = 0;
  if (opa == EVAL_FAIL) {
    // TODO: Should this just return EVAL_FAIL instead, for fwd refs?
    message(pass ? MSG_ERROR : MSG_NOTICE, "symResolveLiteral: symbol '%s' not found",
        alabel);
    if (pass) exit(1); else return EVAL_FAIL;
  } else {
    MESSAGE(MSG_INFO, "symResolveLiteral: Found symbol '%s' = %04x\n", alabel, opa);
  }
  return opa;
}

